#!/usr/bin/perl -w

=head1 NAME

B<POWER::MetaSearch> - Make fast metasearch using many search servers.

=head1 SYNOPSIS

 use POWER::MetaSearch;

 @known_engines = POWER::MetaSearch::engines_list();
 $ms = new POWER::MetaSearch;
 @engines = ("Yahoo","Google");
 $ms->max_results("Google", 50);
 $ms->search("Programming Perl", 40, @engines);
 @yahoo_results = $ms->results("Yahoo");

 $ms->favour("Yahoo", 100);
 $ms->favour("MSN", 75);
 $ms->max_results("Yahoo", 10);
 $ms->max_results("MSN", 20);
 @res = $ms->search("Programming Perl", 30, 
    "Yahoo", "MSN")->by_relevance();

=head1 DESCRIPTION

Fast metasearch.

=head1 KNOWN SEARCH SERVERS

Here the full list of search servers which can be used in metasearch:
'DirectHit', 'MSN', 'AOL', 'AllTheWeb', 'LycosPro', 'Excite',
'NorthernLight', 'Google', 'Yahoo', 'Netscape'.

=head1 METHODS

=over 4

=cut

package POWER::MetaSearch;
use POWER::lib;
use POWER::Multi::GET;
use POWER::HTML::TAG qw(strip);
use strict;
use vars qw($VERSION $DEBUG);
$VERSION = "1.33";
$DEBUG   = 0;

# Define search servers
my %ENGINES = (
	"Google"	=> {
		url		=> "http://www.google.com/search?q=^~query~^&btnG=Google+Search",
#		Rec		=> qr{<p><A HREF=([^>]+)>(.*?)</A>\s*(<font.*?)<font}si,
		Rec		=> qr{<p><A HREF=([^>]+)>(.*?)</A><br>(<font.*?)<font}si,
#		Next	=> qr{<A HREF=(/search\?[^>]+)><IMG [^>]+><br><span [^>]+><b>Next}si,
		Next	=> qr{<A HREF=(/search\?[^>]+)><IMG [^>]+><br><span [^>]+>Next}si,
	    },
	"AllTheWeb"	=> {
		url		=> "http://www.alltheweb.com/search?cat=web&lang=any&query=^~query~^",
#		Rec		=> qr{<dt>.*?href=\"([^\"]*)\">(.*?)</dt>(.*?)<br>}si,
		Rec     => qr{<dt>.*?href=\"([^\"]*)\".*?>(.*?)</dt>(.*?)<br>}si,
		Next	=> qr{<a class=\"link\" href=\"(search\?[^\"]+)\">\&nbsp;\&gt;\&gt;}si,
	    },
	"Excite"	=> {
		url		=> "http://search.excite.com/search.gw?c=web&search=^~query~^&onload=",
		Rec		=> qr{<li>\s+<a\s+href=/r/[^\s]*;([^\s]+).*?>(.*?)<br>(.*?)<br>}si,
		Next	=> qr{<a href=(http://search\.excite\.com/search\.gw\?[^>]+)><b>next}si,
	    },
	"AOL"		=> {
		url		=> "http://search.aol.com/dirsearch.adp?start=&from=topsearchbox.%2Findex.adp&query=^~query~^",
		Rec		=> qr{<li><b>.*?onmouseover=\"self\.status=\'([^\']+)\'[^>]*>(.*?)</A>(.*?)<br>}si,
		Next	=> qr{<A HREF=\"(dirsearch.adp\?[^\"]+)\"><b>more}si,
	    },
	"DirectHit"	=> {
		url		=> "http://www.directhit.com/search.asp?qry=^~query~^&alias=websrch&submit.x=0&submit.y=0",
		Rec		=> qr{\d\. <a href.*?url=(.*?)&.*?<b>(.*?)</a>(.*?)<span}si,
		Next	=> qr{<a href=\"(/umbrella.asp\?[^\"]+)\">Next &gt;&gt;}si,
	    },
	"LycosPro"	=> {
		url		=> "http://lycospro.lycos.com/srchpro/?aloc=sb_init_content&first=1&lpv=1&query=^~query~^&t=all&type=advwebsites",
		Rec		=> qr{<LI><FONT.*?&target=(.*?)&.*?>(.*?)</a>(.*?)</FONT>}si,
		Next	=> qr{<A HREF=(/srchpro/\?[^>]*)><B>next</B></A>}si,
	    },
	"MSN"		=> {
		url		=> "http://search.msn.com/results.asp?FORM=SMCRT&cfg=SMCINITIAL&q=^~query~^&un=doc&v=1",
		Rec		=> qr{<LI /><A CLASS=\"clsResult.*?href=\"(.*?)\">(.*?)</A>(.*?)<br /><}si,
		Next	=> qr{<a CLASS=\"clsNextPrev\"[^>]+href=\"(results.asp\?[^\"]*)\">NEXT}si,
	    },
	"Netscape"	=> {
		url		=> "http://search.netscape.com/search.psp?cp=nsikwphopNetscape&search=^~query~^", 
		Rec		=> qr{<LI><a href=\"http:[^\"]*?/nrpusrws/([^\"]*?)\">(.*?)</a>.*?<br>(.*?)<br>}si,
		Next	=> qr{<a href=\"(/search.psp\?[^\"]*)\">next}si,
	    },
	"NorthernLight"	=> {
		url		=> "http://www.northernlight.com/nlquery.fcg?cb=0&qr=^~query~^&search.x=0&search.y=0",
		Rec		=> qr{NLResultStart.*?<a href=\"([^\"]*)\">(.*?)</a>.*?</b>(.*?)<br>}si,
		Next	=> qr{<a href=\"(/nlquery.fcg\?[^\"]*)\"><img[^>]*Next Page}si,
	    },
	"Yahoo"		=> {
		url		=> "http://google.yahoo.com/bin/query?p=^~query~^&hc=6&hs=166",
		Rec		=> qr{<li>\s*<a href=\"http:[^\"]*?/\*([^\"]*)\">(.*?)</a>(.*?)<font}si,
		Next	=> qr{<a href=\"(/bin/query\?[^\"]*)\">Next}si,
	    },
);
@{$ENGINES{$_}}{"name","favour"} = ($_, 1) for keys %ENGINES;

=item B<new>()

   POWER::MetaSearch.
     POWER::MetaSearch   
  (..    favour, max_results, etc.).
      .

=cut

sub new {
    my $this  = shift ;
    my $class = ref($this) || $this ;
    my $self = {};
    $self->{ENGINES} = { map { $_ => {
	    max_results	=> $this->{ENGINES}{$_}{max_results}, 
    	    favour	=> $this->{ENGINES}{$_}{favour}
	} } keys %{$this->{ENGINES}} } if ref($this);
    bless $self, $class ;
}

=item B<engines_list>()

Return list of supported engines.

=cut

sub engines_list { keys %ENGINES }

=item B<favour>( $engine [, $percent] )

  $percent    $engine.

        - 
        
 $engine.  - -   
   $persent  undef.

    (1-100)     .
       
   .

   (  - 1).

=cut

#      (1-100)
sub _percent { ($_[0]>100) ? 100 : ($_[0]<1) ? 1 : $_[0] }
sub favour {
    my ($self, $engine)  = (shift, shift);
    return undef unless exists $ENGINES{$engine};
    if (ref($self)) {
	$self->{ENGINES}{$engine}{favour} = _percent(shift) if @_;
	return defined($self->{ENGINES}{$engine}{favour}) ?
	    $self->{ENGINES}{$engine}{favour} : $ENGINES{$engine}{favour};
    } else {
	$ENGINES{$engine}{favour} = _percent(shift) if @_;
	return $ENGINES{$engine}{favour};
    }
}

=item B<max_results>( $engine [, $max_results] )

  $max_results    
 $engine.

$max_results       .
    ,    $max_results, 
  .  $max_results undef  
  ,         .

   (  - undef).

=cut

sub max_results {
    my ($self, $engine)  = (shift, shift);
    return undef unless exists $ENGINES{$engine};
    $self->{ENGINES}{$engine}{max_results}=shift if @_;
    return $self->{ENGINES}{$engine}{max_results};
}

=item B<search>( $query, $timeout [, @engines] )

Make metasearch. All results without eliminate duplications and
without sorting will be stored in object.

$query is user query. No special query language used at this time: query
will be sent to search servers as is.

$timeout is timeout in seconds. metasearch finished after downloading asked
number of results for all asked servers, or after timeout.

@engines is a list of engines which will be used in metasearch. If @engines
isn't specified, all known engines will be used.

Return $self (reference to itself) - to use like this: 

 @sorted = $obj->search(...)->by_relevance();

=cut

# Normalize data cutted from html page
sub _norm { local $_ = UnEnc(strip($_[0])); s/\n+/ /g; s/^\s+//; s/\s+$//; $_ }
# Standart parsing: 1) list of urls with title and description, 2) link "Next"
sub _parse {
    my ($url, %Engine, @results, $next_url) = ($_[0], %{$_[1]});
	my $html = $url->{_content};
    push @results, {url => UnEncUri($1), title => _norm($2), descr => _norm($3),
		source_url => $url->{_url}}	while $html=~m!\G.*?$Engine{Rec}!gcsi;
    $next_url=abs_url($1, $Engine{url}) if $html=~m!\G.*?$Engine{Next}!gcsi;
    return (\@results, $next_url);
}
sub search {
    my ($self, $query, $timeout, @engines) = (shift, @_);
    @engines = @engines ? (grep {exists $ENGINES{$_}} @engines) : keys %ENGINES;
    #         
    my %URL = map {
	$self->{ENGINES}{$_}{results} = [];
	(my $url=$ENGINES{$_}{url}) =~ s/\^~query~\^/EncUri($query)/eg;
	($_." - 1", {url=>$url,engine=>$_,from=>1})
	} @engines;
    $timeout = time()+$timeout;
    POWER::Multi::GET::Download(%URL, sub {
	    #       
	    for (grep {$_->{_ok}} values %URL) {
		#    (   timeout)
		delete $_->{_state}, next if $_->{_ok} eq "ERR";
		# DEBUG
		open(DEBUG, ">".$_->{engine}."-".$_->{from}.".html") and
		    print(DEBUG $_->{_header},"\n",$_->{_content}) and 
		    close(DEBUG) or die($!)
		    if $POWER::MetaSearch::DEBUG;
		#  ,  
		my $redirect = $_->{_header} =~ /^Location: (.*?)\r?$/ms;
		my ($results, $next_url) = $redirect ? 
		    ([], abs_url($1, $_->{url})) :
		    _parse($_, $ENGINES{$_->{engine}});
		#       max_results
		my $from = $_->{from} + @$results;
		my $max = $self->max_results($_->{engine});
		$URL{ $_->{engine}." - ".$from } = {
			url	=> $next_url,
			engine	=> $_->{engine},
			from	=> $from,
		    } if $next_url and (not defined $max or $max >= $from);
		#   
		@{$self->{ENGINES}{$_->{engine}}{results}}[$_->{from}-1..
		    $from-2] = @$results;
		delete $URL{$_->{engine}." - ".$_->{from}} unless $redirect;
	    }
	    #  timeout
	    undef(%URL) if $timeout <= time();
	});
    return $self;
}

=item B<results>( $engine )

    $engine ( search*()).

        : url, title, descr 
 source_url. url, title  descr   ; source_url
 url   ,      .

=cut

sub results {
    my ($self, $engine) = (shift, @_);
    return unless exists $self->{ENGINES}{$engine} and
	ref $self->{ENGINES}{$engine}{results} eq "ARRAY";
    #    max_results
    my $max  = $self->max_results($engine);
    my $last = @{$self->{ENGINES}{$engine}{results}};
    $last = $max if defined($max) and $max<$last;
    return @{$self->{ENGINES}{$engine}{results}}[0..$last-1];
}

=item B<by_relevance>( [@engines] )

Return list of sorted results. Sort criteria (relevance) calculated based
on engine favour() and position of result on this engine.
If @engines isn't specified all ready engines are used.
Each result is hash reference with fields 'url', 'title', 'descr',
'engine' and 'relevance'.

=cut

sub by_relevance {
    my ($self, @engines) = (shift, @_);
    @engines = keys %{$self->{ENGINES}} unless @engines;
    my (@results, %seen, $engine, $pos);
    for $engine (@engines) {
	my @r = $self->results($engine);
	for $pos (1..@r) {
	    push @results, { %{$r[$pos-1]}, "engine" => $engine, 
		"relevance" => $self->favour($engine) / $pos };
	}
    }
    return grep {! $seen{$_->{url}}++}
	sort {$b->{relevance} <=> $a->{relevance} or
	    $self->favour($b->{engine}) <=> $self->favour($a->{engine})
	    } @results;
}

=item B<dump_results>( [@engines] )

       
 .    ,  
    .  
    restore_results().

=cut

sub dump_results {
    my ($self, @engines) = (shift, @_);
    @engines = @engines ? (grep {exists $self->{ENGINES}{$_}} @engines) : 
	keys %{$self->{ENGINES}};
    @engines = grep {ref $self->{ENGINES}{$_}{results} eq "ARRAY"} @engines;
    return { map {$_, $self->{ENGINES}{$_}{results}} @engines };
}

=item B<restore_results>( $dump [, @engines] )

  dump_results().
   $dump  .   @engines
    engines  $dump   .

Return $self (reference to itself) - to use like this: 

 @sorted = $obj->restore_results(...)->by_relevance();

=cut

sub restore_results {
    my ($self, $dump, @engines) = (shift, @_);
    @engines = @engines ? (grep {exists $ENGINES{$_}} @engines) : keys %$dump;
    $self->{ENGINES}{$_}{results} = (ref $dump->{$_} eq "ARRAY") ? $dump->{$_}
	: [] for @engines;
    return $self;
}

=back

=head1 SUPPORT (UPDATE SEARCH SERVERS)

 Bes Murzik <murzik@powerman.sky.net.ua>

=head1 AUTHORS

 Alex Efros <powerman@consultant.com>
 Nikita Savin <nikita@only.com>

=cut

1;
